/*
 * Created on 20.08.2003
 *
 * To change this generated comment go to 
 * Window>Preferences>Java>Code Generation>Code Template
 */
package clustering;

import java.util.logging.Level;
import java.util.logging.Logger;

import xxl.core.functions.Function;

/**
 * @author beringer
 */
public abstract class AbstractFuzzyKMeans implements FuzzyKMeansClustering {

    static Logger logger = Logger.getLogger("Test");

    static public FuzzyValidity SEPARATION = new FuzzyValidity() {
        public double invoke(FuzzyKMeansState state) {
            AbstractFuzzyKMeans fkm = (AbstractFuzzyKMeans) state.fkmc;
            double separation = 0d;
            double minDist = Double.MAX_VALUE;
            int c = 0;
            for (int i = 0; i < state.clusterCount-1; i++) {
                for (int j = i + 1; j < state.clusterCount; j++) {
                    double dist = fkm.distance.squaredDistance(state.kMeans[i],
                            state.kMeans[j]);
                    c++;
                    if (dist < minDist)
                        minDist = dist;
                }
            }
            double[] sumCl = new double[state.clusterCount];
            for (int i = 0; i < state.dataCount; i++) {
                double sum = 0.0;
                for (int j = 0; j < state.clusterCount; j++) {
                    double val = state.fuzzyCluster[i][j];
                    sum += val;
                    // val=val*val;
                    if (fkm.fuzzifier == 2)
                        val = val * val;
                    else
                        val = Math.pow(val, fkm.fuzzifier);

                    sumCl[j] += val;

                    double dist = state.clusterSquaredDist[i][j];
                    double val2 = dist * val;
                    separation += val2;
                }
                if (fkm instanceof FuzzyKMeans) {
                    if (((FuzzyKMeans) fkm).noiseClustering) {
                        double val = 1.0 - sum;
                        if (val < 0.0)
                            val = 0.0;
                        if (fkm.fuzzifier == 2)
                            val = val * val;
                        else
                            val = Math.pow(val, fkm.fuzzifier);

                        double dist = ((FuzzyKMeans) fkm).squaredNoiseDistance;
                        double val2 = dist * val;
                        separation += val2;

                    }
                }
            }
            state.sumDist=separation;
            double val2 = minDist;
            if (val2 < 0.00000000001)
                separation = Double.MAX_VALUE;
            else
                separation = separation / (state.dataCount * val2);
            state.validity = separation;

            return separation;
        }
    };

    static public FuzzyValidity CS = new FuzzyValidity() {
        public double invoke(FuzzyKMeansState state) {
            AbstractFuzzyKMeans fkm = (AbstractFuzzyKMeans) state.fkmc;
            double result = 0d;
            double[] maxDist = new double[state.clusterCount];
            double[] minDist = new double[state.clusterCount];
            for (int i = 0; i < state.dataCount; i++) {
                for (int j = i + 1; j < state.dataCount; j++) {
                    double dist = fkm.distance.distance(state.data[i],
                            state.data[j]);
                    for (int c = 0; c < state.clusterCount; c++) {
                        double val = state.fuzzyCluster[i][c]
                                * state.fuzzyCluster[j][c] * dist;
                        if (dist > maxDist[c]) {
                            maxDist[c] = dist;
                        }
                        minDist[c] = Double.MAX_VALUE;
                        for (int c2 = 0; c2 < state.clusterCount; c2++) {
                            if (c2 == c)
                                continue;
                            double distc = fkm.distance.distance(
                                    state.kMeans[c], state.kMeans[c2]);
                            if (distc < minDist[c]) {
                                minDist[c] = distc;
                            }
                        }
                    }
                }
            }
            double num = 0;
            double denom = 0;
            for (int c = 0; c < state.clusterCount; c++) {
                num += maxDist[c];
                denom += minDist[c];
            }
            result = num / denom;
            state.validity = result;
            return result;
        }
    };
    
    static double MAX_VALUE=Math.pow(2,40);

    static public FuzzyValidity MYVF3 = new FuzzyValidity() {
        
        public double invoke(FuzzyKMeansState state) {
            Distance distance = state.fkmc.getDistance();
            double minClusterDist = 0;
            double[] sumC = new double[state.clusterCount];
            double[] minC = new double[state.clusterCount];
            double[] sumD = new double[state.clusterCount];
            int[] countC = new int[state.clusterCount];
            double penalty=MAX_VALUE*0.01*state.clusterCount;
            double sum = 0;
            double fuzzifier = state.fkmc.getFuzzifier();
            for (int i = 0; i < state.dataCount; i++) {
                for (int c = 0; c < state.clusterCount; c++) {
                    double val = 0;
                    if (fuzzifier == 2.0)
                        val = state.fuzzyCluster[i][c] * state.fuzzyCluster[i][c];
                    else
                        val = Math.pow(state.fuzzyCluster[i][c], fuzzifier);
                    double dist = state.clusterSquaredDist[i][c] * val;
                    sumD[c] += dist;
                    sumC[c] += val;
                    sum += dist;
                    if (state.fuzzyCluster[i][c] > 0.5)
                        countC[c]++;
                }
            }
            state.sumDist=sum;
            for (int c = 0; c < state.clusterCount; c++) {
                sumD[c]=Math.sqrt(sumD[c] / sumC[c]);
            }
            for (int c = 0; c < state.clusterCount - 1; c++) {
                for (int c2 = c + 1; c2 < state.clusterCount; c2++) {
                    double distc = distance.squaredDistance(state.kMeans[c], state.kMeans[c2]);
                    minClusterDist += 1 / (distc / (sumD[c]*sumD[c2]));
                    minC[c] += distc;
                    minC[c2] += distc;
                }
            }
            double result = 0;
            double[] valCl = new double[state.clusterCount];
            int small = 0;
            for (int c = 0; c < state.clusterCount; c++) {
                if (countC[c] < 4)
                    small++;
                else if (minC[c] == 0) {
                    if (state.clusterCount > 1)
                        small++;
                }
            }
            
            if (Double.isInfinite(minClusterDist)|| minClusterDist>penalty*0.001) {
                minClusterDist = penalty*0.001;
            }
            
            result = sum * minClusterDist / state.dataCount;
            result += small * penalty;
            if (Double.isInfinite(result)|| result>MAX_VALUE) {
                result = MAX_VALUE;
            }
            state.validity = result;
            logger.log(Level.FINER, state.clusterCount + "     " + result + "\n\n");
            return result;
        }
    };

    

    
    static public FuzzyValidity FVF = new FuzzyValidity() {
        public double invoke(FuzzyKMeansState state) {
            Distance distance = state.fkmc.getDistance();
            double[] sumC = new double[state.clusterCount];
            double[][] sumComb = new double[state.clusterCount][state.clusterCount];
            double sumHet=0;
            int[] countC = new int[state.clusterCount];
            double sum = 0;
            double val=0;
            double val2=0;
            double fuzzifier = state.fkmc.getFuzzifier();
            
            for (int i = 0; i < state.dataCount; i++) {
                
                for (int c = 0; c < state.clusterCount; c++) {
                    if (fuzzifier == 2.0)
                        val = state.fuzzyCluster[i][c]
                                * state.fuzzyCluster[i][c];
                    else
                        val = Math.pow(state.fuzzyCluster[i][c], fuzzifier);
                    double dist = state.clusterSquaredDist[i][c] * val;
                    sumC[c] += val;
                    sum += dist;
                    if (state.fuzzyCluster[i][c] > 0.5)
                        countC[c]++;
                    for (int c2 = c + 1; c2 < state.clusterCount; c2++) {
                        if (fuzzifier == 2.0)
                            val2 = state.fuzzyCluster[i][c2]
                                    * state.fuzzyCluster[i][c2];
                        else
                            val2 = Math.pow(state.fuzzyCluster[i][c2], fuzzifier);
                        sumComb[c2][c]+=val*val2;
                    }
                }
            }
            state.sumDist=sum;
            
            for (int c = 0; c < state.clusterCount - 1; c++) {
                for (int c2 = c + 1; c2 < state.clusterCount; c2++) {
                    double het1=sumComb[c2][c]/sumC[c];
                    double het2=sumComb[c2][c]/sumC[c2];
                    sumHet=sumHet+het1-sumHet*het1;
                    sumHet=sumHet+het2-sumHet*het2;
                }
            }
            double result = 0;
            double[] valCl = new double[state.clusterCount];
            int small = 0;
            for (int c = 0; c < state.clusterCount; c++) {
                if (countC[c] < 3)
                    small++;
                
            }
            
            result = sum / ((1-sumHet)* state.dataCount);
            result += small * Double.MAX_VALUE / state.clusterCount;
            if (Double.isInfinite(result)) {
                result = Double.MAX_VALUE;
            }
            state.validity = result;
            logger.log(Level.FINER, state.clusterCount + "     " + result + "\n\n");
            return result;
        }
    };

    static public FuzzyValidity DIM_VF = new FuzzyValidity() {
        public double invoke(FuzzyKMeansState state) {
            DimDistance distance = (DimDistance)state.fkmc.getDistance();
            int dimCount=distance.getDimensionCount();
            double minClusterDist = 0;
            double[] sumC = new double[state.clusterCount];
            double[] minC = new double[state.clusterCount];
            double[][] sumD = new double[state.clusterCount][dimCount];
            
            double[][][] clusterDimDist=new double[state.dataCount][state.clusterCount][dimCount];
            
            int[] countC = new int[state.clusterCount];
            double sum = 0;
            double fuzzifier = state.fkmc.getFuzzifier();
            for (int i = 0; i < state.dataCount; i++) {
                for (int c = 0; c < state.clusterCount; c++) {
                    double val = 0;
                    if (fuzzifier == 2.0)
                        val = state.fuzzyCluster[i][c] * state.fuzzyCluster[i][c];
                    else
                        val = Math.pow(state.fuzzyCluster[i][c], fuzzifier);
                    double dist = state.clusterSquaredDist[i][c] * val;
                    sumC[c] += val;
                    sum += dist;
                    if (state.fuzzyCluster[i][c] > 0.5)
                        countC[c]++;
                    for(int d=0;d<dimCount;d++){
                        double dimDist=distance.squaredDistance(state.data[i],state.kMeans[c],d);
                        clusterDimDist[i][c][d]=dimDist;
                        sumD[c][d]+=Math.sqrt(dimDist)*val;
                    }
                }
            }
            state.sumDist=sum;
            for (int c = 0; c < state.clusterCount; c++) {
                for(int d=0;d<dimCount;d++){
                    sumD[c][d]=sumD[c][d]/sumC[c];
                }
            }
            for (int c = 0; c < state.clusterCount - 1; c++) {
                for (int c2 = c + 1; c2 < state.clusterCount; c2++) {
                    double distc = distance.squaredDistance(state.kMeans[c], state.kMeans[c2]);
                    double meanDist=0;
                    double meanDist2=0;
                    for(int d=0;d<dimCount;d++){
                        double dimDist=distance.squaredDistance(state.kMeans[c], state.kMeans[c2],d);
                        
                        meanDist+=(sumD[c][d] )*dimDist/distc;
                        meanDist2+=(sumD[c2][d] )*dimDist/distc;
                    }
                    distc=Math.sqrt(distc);
                    minClusterDist += 1 / (distc / (meanDist + meanDist2));
                    minC[c] += distc;
                    minC[c2] += distc;
                }
            }
            double result = 0;
            double[] valCl = new double[state.clusterCount];
            int small = 0;
            for (int c = 0; c < state.clusterCount; c++) {
                if (countC[c] < 3)
                    small++;
                else if (minC[c] == 0) {
                    if (state.clusterCount > 1)
                        small++;
                }
            }
            result = sum * minClusterDist / state.dataCount;
            result += small * Double.MAX_VALUE / state.clusterCount;
            if (Double.isInfinite(result)) {
                result = Double.MAX_VALUE;
            }
            state.validity = result;
            logger.log(Level.FINER, state.clusterCount + "     " + result + "\n\n");
            return result;
        }
    };
    
    
    

    protected Distance distance;

    /** Fuzzifierconstant (>1), Defaultvalue=2 */
    protected double fuzzifier;

    protected FuzzyValidity validity;

    protected FuzzyObjective objective;

    protected Initialization initialize;

    protected Function copyObject;

    protected FuzzyAverage meansFunction;

    public AbstractFuzzyKMeans(Distance distance, double fuzzifier,
            FuzzyObjective objective, FuzzyValidity validity,
            Initialization initialize, Function copyObject,
            FuzzyAverage meansFunction) {
        this.fuzzifier = fuzzifier;
        this.distance = distance;
        this.validity = validity;
        this.objective = objective;
        this.initialize = initialize;
        this.copyObject = copyObject;
        this.meansFunction = meansFunction;
        // System.out.println("Fuzzifier: "+fuzzifier);
    }

    public Distance getDistance() {
        return distance;
    }

    public double getFuzzifier() {
        return fuzzifier;
    }

    /**
     * compute weighted means of the clusters
     * 
     * @param obs
     *            Objects
     * @param len
     *            Objcetcount
     * @param cluster
     *            membership degrees
     * @param clusterNr
     *            cluster number to compute center
     * @param means
     *            Object to store center, if null a new Object has to be created
     * @return the means
     */
    protected Object computeMeans(Object[] obs, int len, double[][] cluster,
            int clusterNr, Object means) {
        return meansFunction.average(this, obs, len, cluster, clusterNr, means);
    }

    public void initialize(FuzzyKMeansState state) {
        initialize.invoke(state);
    }

    public FuzzyKMeansState invokeFuzzyKMeans(Object[] data, int dataCount,
            int clusterCount, double epsilon) {
        Object[] kmeans = new Object[clusterCount];

        double[][] clusterDist = new double[dataCount][clusterCount];
        double[] clusterCounts = new double[clusterCount];
        double[][] cluster = new double[dataCount][clusterCount];
        FuzzyKMeansState state = new FuzzyKMeansState(data, dataCount,
                clusterCount);
        state.epsilon = epsilon;
        state.clusterSquaredDist = clusterDist;
        state.clusterCounts = clusterCounts;
        state.fuzzyCluster = cluster;
        state.maxClusterCount = clusterCount;
        state.minClusterCount = clusterCount;
        state.fkmc = this;
        state.kMeans = kmeans;
        initialize(state);
        computeMemberships(state);
        double lastObjective = computeObjective(state);
        double oldQual = lastObjective;
        do {
            computeClusterCenters(state);
            computeMemberships(state);
            oldQual = lastObjective;
            lastObjective = computeObjective(state);
        } while (Math.abs(lastObjective - oldQual) > epsilon);
        computeValidity(state);
        return state;
    }

    public double[] computeMembership(Object[] kmeans, int clusterCount,
            Object ob) {
        return computeMembership(kmeans, clusterCount, new Object[] { ob })[0];
    }

    public double[][] computeMembership(Object[] kmeans, int clusterCount,
            Object[] ob) {
        int dataCount = ob.length;
        double[][] cluster = new double[dataCount][clusterCount];
        double[] clusterCounts = new double[clusterCount];
        double[][] clusterSquaredDist = new double[dataCount][clusterCount];
        FuzzyKMeansState state = new FuzzyKMeansState();
        state.clusterCount = clusterCount;
        state.epsilon = -1;
        state.clusterSquaredDist = clusterSquaredDist;
        state.clusterCounts = clusterCounts;
        state.fuzzyCluster = cluster;
        state.maxClusterCount = clusterCount;
        state.minClusterCount = clusterCount;
        state.fkmc = this;
        state.kMeans = kmeans;
        computeMemberships(state);
        return cluster;
    }

    public void computeClusterCenters(FuzzyKMeansState state) {
        for (int i = 0; i < state.clusterCount; i++) {
            state.kMeans[i] = computeMeans(state.data, state.dataCount,
                    state.fuzzyCluster, i, state.kMeans[i]);
        }

    }

    public double computeValidity(FuzzyKMeansState state) {

        return validity.invoke(

        state);
    }

    public double computeObjective(FuzzyKMeansState state) {
        return ((Double) objective.invoke(

        state)).doubleValue();
    }

    public Function getCopyFunction() {
        return copyObject;
    }

}
